import logging
from typing import List
from bson import SON

from pymongo import IndexModel, ASCENDING

from ..connection import get_db
from ..models import Document, INHERITANCE_FIELD_NAME

logger = logging.getLogger(__name__)


def index_for_a_collection(operation):
    db = get_db()
    """
    First get all incexes for a collection and match with operation.
    Remove full match object.

    If db_index partially match with operation_index then recreate/update it.

    For new_indexes unmatch with db_indexes create new index.
    For db_indexes unmatch with new_indexes drop indexes.
    """
    try:
        collection = db[operation["collection_name"]]
        indexes = operation["create_indexes"]
    except Exception:
        raise Exception("Invalid index object")

    # print(indexes)
    db_indexes = []
    for index in collection.list_indexes():
        temp_val = index.to_dict()  # type: ignore
        # Skip "_id" index since it's create by mongodb system
        if "_id" in temp_val["key"]:
            continue
        temp_val.pop("v", None)
        db_indexes.append(temp_val)

    new_indexes = []
    new_indexes_store = {}

    for index in indexes:
        temp_val = index.document
        # Replace SON object with dict
        if type(temp_val["key"]) == SON:
            temp_val["key"] = temp_val["key"].to_dict()
        else:
            continue
        new_indexes.append(temp_val)
        # Store index object for future use
        new_indexes_store[temp_val["name"]] = index

    # print(db_indexes)
    # print(new_indexes)

    update_indexes = []
    for i in range(len(db_indexes)):
        partial_match = None
        for j in range(len(new_indexes)):
            if type(new_indexes[j]) is not dict:
                continue
            if db_indexes[i] == new_indexes[j]:
                db_indexes[i], new_indexes[j] = None, None
                partial_match = None
                break

            """
            # TODO: make a list for partial match
            if pertial match db_indexes[i] with new_indexes[i]:
                parial_match = j
                # not break here check if any other match exist
            """

        if partial_match is not None:
            update_indexes.append((db_indexes[i], new_indexes[partial_match]))
            db_indexes[i], new_indexes[partial_match] = None, None

    delete_db_indexes = [val for val in db_indexes if val]
    new_indexes = [val for val in new_indexes if val]

    for db_index in delete_db_indexes:
        if db_index is not None:
            collection.drop_index(db_index["name"])
    if len(new_indexes) > 0:
        new_indexes = [
            new_indexes_store[new_index["name"]] for new_index in new_indexes if new_index
        ]
        try:
            collection.create_indexes(new_indexes)
        except Exception as e:
            print(f'\nProblem arise at "{operation["collection_name"]}": {e}\n')
            raise Exception(str(e))

    # TODO: apply action for update_indexes

    ne, de = len(new_indexes), len(delete_db_indexes)
    if ne > 0 or de > 0:
        logger.info(
            f'Applied for "{operation["collection_name"]}": {de} deleted, {ne} added'
        )
    return ne, de


def get_model_indexes(model) -> List:
    if hasattr(model.Config, "indexes"):
        return list(model.Config.indexes)
    return []


def get_all_indexes():
    """
    First imports all child models of Document since it's the abstract parent model.
    Then retrieve all the child modules and will try to get indexes inside the Config class.
    """
    operations = []
    for model in Document.__subclasses__():
        indexes = get_model_indexes(model)
        if indexes:
            collection_name = model._get_collection_name()
            obj = {
                "collection_name": collection_name,
                "create_indexes": indexes,
            }
            if (
                hasattr(model.Config, "allow_inheritance")
                and model.Config.allow_inheritance is True
            ):
                """If a model has child model"""
                indexes.append(IndexModel([(INHERITANCE_FIELD_NAME, ASCENDING)]))
                for child_model in model.__subclasses__():
                    indexes += get_model_indexes(child_model)
            operations.append(obj)
    return operations


def apply_indexes():
    """Run "python -m app.main applyindexes" to apply and indexes."""

    """First get all indexes from all model."""
    operations = get_all_indexes()

    """Then excute each indexes operation for each model."""
    new_index, delete_index = 0, 0
    for operation in operations:
        ne, de = index_for_a_collection(operation)
        new_index += ne
        delete_index += de

    if delete_index:
        logger.info(f"{delete_index}, index deleted.")
    if new_index:
        logger.info(f"{new_index}, index created.")
    if [new_index, delete_index] == [0, 0]:
        logger.info("No change detected.")



import logging
from datetime import datetime
from typing import Any, Iterator, List, Optional, Tuple
from typing_extensions import Self
from bson import ObjectId


from pydantic import BaseModel, Field
from pymongo import DESCENDING
from pymongo.cursor import Cursor
from pymongo.results import UpdateResult, DeleteResult
from pymongo.collection import Collection

from .connection import get_db
from .data_conversion import dict2obj
from .types import PydanticObjectId  # type: ignore
from .utils.utils import convert_model_to_collection


INHERITANCE_FIELD_NAME = "_cls"
logger = logging.getLogger(__name__)


class _BaseDocument(BaseModel):
    class Config:
        # Those fields will work as the default value of any child class.
        orm_mode = True
        allow_population_by_field_name = True
        collection_name = None
        allow_inheritance = False

    def __init__(self, *args, **kwargs) -> None:
        if type(self) is Document:
            raise Exception(
                "Document is an abstract class and cannot be instantiated directly"
            )
        super().__init__(*args, **kwargs)

    def __setattr__(self, key, value) -> None:
        """Add '# type: ignore' as a comment if get type error while getting this value"""
        self.__dict__[key] = value

    @classmethod
    def _get_collection_class(cls) -> Any:
        model: Any = cls
        if model.__base__ != Document:
            base_model = model.__base__
            if (
                not hasattr(base_model.Config, "allow_inheritance")
                or base_model.Config.allow_inheritance is not True
            ):
                raise Exception(
                    f"Invalid model inheritance. {base_model} does not allow model inheritance."
                )
            if base_model.Config == model.Config:
                raise Exception(
                    f"Child Model{model.__name__} should declare a separate Config class."
                )
            return base_model, model
        else:
            return model, None

    @classmethod
    def _get_collection_name(cls) -> str:
        collection, _ = cls._get_collection_class()
        return convert_model_to_collection(collection)

    @classmethod
    def _get_child(cls) -> Optional[str]:
        _, collection = cls._get_collection_class()
        if collection is None:
            return None
        return convert_model_to_collection(collection)

    @classmethod
    def _get_collection(cls) -> Collection:
        if not hasattr(cls, "_collection") or cls._collection is None:
            db = get_db()
            cls._collection = db[cls._get_collection_name()]
        return cls._collection

    @classmethod
    def _db(cls) -> str:
        return cls._get_collection_name()

    # @classmethod
    # def start_session(cls):
    #     return get_client().start_session()


class Document(_BaseDocument):
    id: PydanticObjectId = Field(default_factory=ObjectId, alias="_id")

    def create(self, **kwargs) -> Self:
        _collection = self._get_collection()

        data = self.dict(exclude={"id"})
        if self._get_child() is not None:
            data = {f"{INHERITANCE_FIELD_NAME}": self._get_child(), **data}

        inserted_id = _collection.insert_one(data, **kwargs).inserted_id
        self.__dict__.update({"id": inserted_id})
        return self

    @classmethod
    def find_raw(cls, filter: dict = {}, projection: dict = {}, **kwargs) -> Cursor:
        _collection = cls._get_collection()
        if cls._get_child() is not None:
            filter = {f"{INHERITANCE_FIELD_NAME}": cls._get_child(), **filter}
        if projection:
            return _collection.find(filter, projection, **kwargs)
        return _collection.find(filter, **kwargs)

    @classmethod
    def find(
        cls,
        filter: dict = {},
        sort: Optional[List[Tuple[str, int]]] = None,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        **kwargs,
    ) -> Iterator[Self]:
        qs = cls.find_raw(filter, **kwargs)
        if sort:
            qs = qs.sort(sort)
        if skip:
            qs = qs.skip(skip)
        if limit:
            qs = qs.limit(limit)

        model_childs = {}
        is_dynamic_model = False
        if (
            hasattr(cls.Config, "allow_inheritance")
            and cls.Config.allow_inheritance is True
        ):
            is_dynamic_model = True
            for model in cls.__subclasses__():
                model_childs[cls._get_child()] = model

        for data in qs:
            if is_dynamic_model and data.get(INHERITANCE_FIELD_NAME) in model_childs:
                yield model_childs[data[INHERITANCE_FIELD_NAME]](**data)
            else:
                yield cls(**data)

    @classmethod
    def __find_one(
        cls, filter: dict = {}, sort: Optional[List[Tuple[str, int]]] = None, **kwargs
    ) -> Optional[Self]:
        qs = cls.find_raw(filter, **kwargs)
        if sort:
            qs = qs.sort(sort)
        for data in qs.limit(1):
            """limit 1 is equivalent to find_one and that is implemented in pymongo find_one"""
            return cls(**data)
        return None

    @classmethod
    def get_or_create(
        cls, filter: dict = {}, session=None, **kwargs
    ) -> Tuple[Self, bool]:
        obj = cls.__find_one(filter)
        if obj:
            return obj, False
        return cls(**filter).create(session=session, **kwargs), True

    @classmethod
    def get(
        cls, filter: dict = {}, sort: Optional[List[Tuple[str, int]]] = None, **kwargs
    ) -> Self:
        obj = cls.__find_one(filter, sort=sort, **kwargs)
        if obj:
            return obj
        raise Exception("Object not found.")

    @classmethod
    def find_first(
        cls, filter: dict = {}, sort: Optional[List[Tuple[str, int]]] = None, **kwargs
    ) -> Optional[Self]:
        return cls.__find_one(filter, sort=sort, **kwargs)

    @classmethod
    def find_last(
        cls,
        filter: dict = {},
        sort: Optional[List[Tuple[str, int]]] = [("_id", DESCENDING)],
        **kwargs,
    ) -> Optional[Self]:
        return cls.__find_one(filter, sort=sort, **kwargs)

    @classmethod
    def count_documents(cls, filter: dict = {}, **kwargs) -> int:
        _collection = cls._get_collection()
        if cls._get_child() is not None:
            filter = {f"{INHERITANCE_FIELD_NAME}": cls._get_child(), **filter}
        return _collection.count_documents(filter, **kwargs)

    @classmethod
    def exists(cls, filter: dict = {}, **kwargs) -> bool:
        _collection = cls._get_collection()
        if cls._get_child() is not None:
            filter = {f"{INHERITANCE_FIELD_NAME}": cls._get_child(), **filter}
        return _collection.count_documents(filter, **kwargs, limit=1) >= 1

    @classmethod
    def aggregate(cls, pipeline: List[Any], _raw=False, **kwargs) -> Iterator[Any]:
        _collection = cls._get_collection()
        if cls._get_child() is not None:
            pipeline = [
                {"$match": {f"{INHERITANCE_FIELD_NAME}": cls._get_child()}}
            ] + pipeline
        for obj in _collection.aggregate(pipeline, **kwargs):
            if _raw is True:
                yield obj
            else:
                yield dict2obj(obj)

    @classmethod
    def get_random_one(cls, filter: dict = {}, **kwargs) -> Optional[Self]:
        _collection = cls._get_collection()
        if cls._get_child() is not None:
            filter[INHERITANCE_FIELD_NAME] = cls._get_child()
        pipeline = [{"$match": filter}, {"$sample": {"size": 1}}]
        for data in _collection.aggregate(pipeline, **kwargs):
            return cls(**data)
        return None

    def update(self, raw: dict = {}, **kwargs) -> UpdateResult:
        _collection = self._get_collection()
        filter = {"_id": self.id}
        if raw:
            updated_data = raw
        else:
            updated_data = {"$set": self.dict(exclude={"id"})}
        if hasattr(self, "updated_at"):
            datetime_now = datetime.utcnow()
            if "$set" not in updated_data:
                updated_data["$set"] = {}
            updated_data["$set"]["updated_at"] = datetime_now
            self.__dict__.update({"updated_at": datetime_now})

        return _collection.update_one(filter, updated_data, **kwargs)

    @classmethod
    def update_one(cls, filter: dict = {}, data: dict = {}, **kwargs) -> UpdateResult:
        _collection = cls._get_collection()
        if cls._get_child() is not None:
            filter = {f"{INHERITANCE_FIELD_NAME}": cls._get_child(), **filter}
        return _collection.update_one(filter, data, **kwargs)

    @classmethod
    def update_many(cls, filter: dict = {}, data: dict = {}, **kwargs) -> UpdateResult:
        _collection = cls._get_collection()
        if cls._get_child() is not None:
            filter = {f"{INHERITANCE_FIELD_NAME}": cls._get_child(), **filter}
        return _collection.update_many(filter, data, **kwargs)

    def delete(self, **kwargs) -> DeleteResult:
        _collection = self._get_collection()
        return _collection.delete_one({"_id": self.id}, **kwargs)

    @classmethod
    def delete_many(cls, filter: dict = {}, **kwargs) -> DeleteResult:
        _collection = cls._get_collection()
        if cls._get_child() is not None:
            filter = {f"{INHERITANCE_FIELD_NAME}": cls._get_child(), **filter}
        return _collection.delete_many(filter, **kwargs)



import re
from typing import Any

pattern = re.compile(r"(?<!^)(?=[A-Z])")


def camel_to_snake(string: str) -> str:
    return pattern.sub("_", string).lower()


def convert_model_to_collection(model: Any) -> str:
    if (
        hasattr(model.Config, "collection_name")
        and model.Config.collection_name is not None
    ):
        """By default model has Config in Basemodel"""
        return model.Config.collection_name
    else:
        return camel_to_snake(model.__name__)



from bson import ObjectId


class PydanticObjectId(ObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, ObjectId):
            return v
        elif isinstance(v, str):
            return ObjectId(v)
        raise TypeError("Invalid data. ObjectId required")



from pymongo import MongoClient
from pymongo.database import Database

__connection = {}


def _get_connection_client(url: str) -> MongoClient:
    return MongoClient(url)


def connect(url: str) -> MongoClient:
    if "client" in __connection:
        return __connection["client"]
    __connection["url"] = url
    __connection["client"] = _get_connection_client(url)
    return __connection["client"]


def disconnect() -> bool:
    if "client" in __connection:
        __connection["client"].close()
        del __connection["client"]
    return True


def force_disconnect() -> bool:
    disconnect()
    if "url" in __connection:
        del __connection["url"]
    return True


def get_client() -> MongoClient:
    if not __connection or "client" not in __connection:
        if "url" in __connection:
            connect(__connection["url"])
        else:
            raise Exception("DB connection is not provided")
    return __connection["client"]


def get_db() -> Database:
    return get_client().get_database()



from typing import Any


class OdmObj(object):
    def __repr__(self) -> str:
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, other) -> bool:
        return self.__dict__ == other.__dict__

    def __iter__(self):
        yield from self.__dict__.items()

    def dict(self):
        return _to_dict(self)


def _to_dict(obj):
    if isinstance(obj, list):
        obj = [_to_dict(x) for x in obj]
    if not isinstance(obj, OdmObj):
        return obj

    n_d = {}
    for k, v in obj:
        n_d[k] = _to_dict(v)
    return n_d


def _to_obj(d):
    if isinstance(d, list):
        d = [_to_obj(x) for x in d]
    if not isinstance(d, dict):
        return d

    obj = OdmObj()
    for k, v in d.items():
        obj.__dict__[k] = _to_obj(v)
    return obj


def dict2obj(d: dict) -> Any:
    return _to_obj(d)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mongodb_odm', 'mongodb_odm.utils']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.10.0,<2.0.0', 'pymongo>=4.3.3,<5.0.0']

setup_kwargs = {
    'name': 'mongodb-odm',
    'version': '0.1.0a3',
    'description': 'MongoDB-ODM, NOSQL databases in Python, designed for simplicity, compatibility, and robustness.',
    'long_description': '# ODM\n\n<p align="center">\n    <em>MongoDB-ODM, NOSQL databases in Python, designed for simplicity, compatibility, and robustness.</em>\n</p>\n\n<p align="center">\n\n<a href="https://github.com/nayan32biswas/mongodb-odm" target="_blank">\n    <img alt="GitHub all releases" src="https://img.shields.io/github/downloads/nayan32biswas/mongodb-odm/total?color=success">\n</a>\n<a href="https://pypi.org/project/mongodb-odm/">\n    <img alt="PyPI" src="https://img.shields.io/pypi/v/mongodb-odm?color=blue">\n</a>\n</p>\n\n---\n\n## Introduction\n\nThe purpose of this module is to do provide easy access to the database with the python object feature with MongoDB and pymongo. With pymongo that was very easy to make spelling mistakes of a collection name when you are doing database operation. This module provides you minimal ODM with a modeling feature so that you don’t have to look up the MongoDB dashboard(Mongo Compass) to know about field names or data types.\n\n**MongoDb-ODM** is based on Python type annotations, and powered by <a href="https://pymongo.readthedocs.io/en/stable/" class="external-link" target="_blank">PyMongo</a> and <a href="https://docs.pydantic.dev/" class="external-link" target="_blank">Pydantic</a>.\n\nThe key features are:\n\n- **Intuitive to write**: Great editor support. Completion everywhere. Less time debugging. Designed to be easy to use and learn. Less time reading docs.\n- **Easy to use**: It has sensible defaults and does a lot of work underneath to simplify the code you write.\n- **Compatible**: It is designed to be compatible with **FastAPI**, **Pydantic**, and **PyMongo**.\n- **Extensible**: You have all the power of **PyMongo** and **Pydantic** underneath.\n- **Short**: Minimize code duplication. A single type annotation does a lot of work. No need to duplicate models in **PyMongo** and Pydantic.\n\n---\n\n## Requirement\n\n**MongoDb-ODM** will work on <a href="https://www.python.org/downloads/" class="external-link" target="_blank">Python 3.7 and above</a>\n\nThis **MongoDb-ODM** is build top of **PyMongo** and **Pydantic**. Those package are required and will auto install while **MongoDb-ODM** was installed.\n\n## Example\n\n#### Define model\n\n```py\nfrom datetime import datetime\nfrom pydantic import Field\nfrom pymongo import IndexModel, ASCENDING\nfrom typing import Optional\n\nfrom mongodb_odm import Document\n\n\nclass User(Document):\n    username: str = Field(...)\n    email: Optional[str] = Field(default=None)\n    full_name: str = Field(...)\n\n    is_active: bool = True\n    date_joined: datetime = Field(default_factory=datetime.utcnow)\n\n    last_login: datetime = Field(default_factory=datetime.utcnow)\n    password: Optional[str] = Field(default=None)\n    image: Optional[str] = Field(default=None)\n\n    created_at: datetime = Field(default_factory=datetime.utcnow)\n    updated_at: datetime = Field(default_factory=datetime.utcnow)\n\n    class Config:\n        collection_name = "user"\n        indexes = (\n            IndexModel([("username", ASCENDING)], unique=True),\n            IndexModel([("email", ASCENDING)]),\n        )\n```\n\n#### Create Document\n\n```py\nuser = User(\n    email="example@example.com",\n    full_name="Example Name",\n    password="hash-password",\n).create()\n```\n\n#### Retrive Document\n\n- Filter data from collection\n\n```py\nfor user in find({"is_active": True}):\n    print(user)\n```\n\n- Find first object with filter\n\n```py\nuser = User.find_first({"is_active": True})\n```\n\n#### Update Data\n\n\n```py\nuser = User.find_first({"is_active": True})\nuser.full_name = "New Name"\n```\n\n#### Delete Data\n\n```py\nuser = User.find_first({"is_active": True})\nif user:\n    user.delete()\n```\n\n### Apply Indexes\n\n```py\nfrom mongodb_odm import apply_indexes, ASCENDING, Document, IndexModel\n\n\nclass User(Document):\n    ...\n\n    class Config:\n        indexes = (\n            IndexModel([("username", ASCENDING)], unique=True),\n            IndexModel([("email", ASCENDING)]),\n        )\n```\n\n- To create indexes in database diclare [IndexModel](https://pymongo.readthedocs.io/en/stable/tutorial.html#indexing) and assign in indexes array in Config class. **IndexModel** module that are directly imported from **pymongo**.\n- Call `apply_indexes` function from your CLI. You can use [Typer](https://typer.tiangolo.com/) to implement CLI.\n',
    'author': 'Nayan Biswas',
    'author_email': 'nayan32biswas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/nayan32biswas/mongodb-odm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
